

#include "c8051.h"

#include "sysclk.h"

#include "uart0.h"
#include "queue8.h"
#include "keycode.h"
#include "waitloop.h"

#define NULL 0



#define LED_NUM P0_2
#define LED_CAPS P0_1
#define LED_SCROLL P0_0

#define KBRTS P0_4
#define PS2CLK P0_6
#define PS2DAT P0_7









volatile unsigned char timerevent;

void timer2_handler(void) interrupt 5
{
	TF2H = 0;
	timerevent = 1;
}

void init_timer2(int counts)
{
	unsigned short countsL = -counts;
	unsigned short countsH = -counts;
	countsH >>= 8;

	TMR2CN = 0x00;
	CKCON &= ~0x60;

	TMR2RLL = countsL;
	TMR2RLH = countsH;
	TMR2L = 0xff;
	TMR2H = 0xff;
	ET2 = 1;
	TR2 = 1;
}



struct Queue8 event_queue;
struct Queue8 pc_input_queue;
struct Queue8 pc_output_queue;


void kb_enable(bool f)
{
	KBRTS = f ? 1 : 0;
}

int kb_get()
{
	return uart0_get();
}

void kb_put(unsigned char c)
{
	uart0_put(c);
}

int pc_get()
{
	return qget(&pc_input_queue);
}

void pc_put(unsigned char c)
{
	qput(&pc_output_queue, c);
}


void put_event(unsigned short c)
{
	qput(&event_queue, c);
	qput(&event_queue, c >> 8);
}

unsigned short get_event()
{
	unsigned short l, h;
	l = qget(&event_queue);
	if (l > 255) return 0;
	h = qget(&event_queue);
	return (h << 8) | l;
}



extern bool pc_send(int c);
extern int pc_recv();
extern bool pc_get_clock();
extern void pc_set_clock_0();
extern void pc_set_clock_1();
extern bool pc_get_data();
extern void pc_set_data_0();
extern void pc_set_data_1();


void system_handler()
{
	int c;

	c = pc_recv();
	if (c >= 0) {
		qput(&pc_input_queue, c);
	}

	c = qget(&pc_output_queue);
	if (c >= 0) {
		if (!pc_send(c)) {
			qunget(&pc_output_queue, c);	// failed
		}
	}
}




enum {
	PC_INIT			= 0x0000,
	PC_NORMAL		= 0x0100,
	PC_ED			= 0x0200,
	PC_F0			= 0x0300,
	PC_F3			= 0x0400,
	PC_FB			= 0x0500,
	PC_FC			= 0x0600,
	PC_FD			= 0x0700,
	PC_FF			= 0x0800,
};

enum {
	EVENT_NULL = 0,
	EVENT_INIT		= 0x0100,
	EVENT_KEYMAKE		= 0x0200,
	EVENT_KEYBREAK		= 0x0300,
	EVENT_SEND_KB_INDICATOR	= 0x0400,
	EVENT_SEND_KB_TYPEMATIC	= 0x0500,
};


unsigned short pc_state;
//unsigned short kb_state;
unsigned short shiftflags;
unsigned char pc_init_count;
unsigned char indicator;
unsigned short typematic;
unsigned short repeat_count;
unsigned short repeat_event;
//unsigned short real_keyboard_id;
//unsigned short fake_keyboard_id;
unsigned char kb_break_delay;
unsigned short kb_break_event;
//bool break_enable;
//bool scan_enable;
#define fake_keyboard_id 0x83ab
#define break_enable 1
#define scan_enable 1

code static const unsigned short typematic_delay_table[] = {
	250,
	500,
	750,
	1000,
};
code static const unsigned short typematic_rate_table[] = {
	33,
	37,
	42,
	46,
	50,
	54,
	58,
	63,

	67,
	75,
	83,
	91,
	100,
	109,
	116,
	125,

	133,
	149,
	167,
	182,
	200,
	217,
	233,
	250,

	270,
	303,
	333,
	370,
	400,
	235,
	476,
	500,
};

unsigned short minuteevent;
unsigned char kbrefresh;

void handler()
{
	int c;
	unsigned short event;

	if (timerevent) {			// 1ms interval
		timerevent = false;

		if (pc_init_count) {
			if (!--pc_init_count) {
				pc_put(0xaa);
			}
		}
		if (kb_break_delay > 0) {
			if (!--kb_break_delay) {
				if (break_enable) {
					put_event(kb_break_event);
				}
				if ((kb_break_event & 0xff) == (repeat_event & 0xff)) {
					repeat_event = 0;
					repeat_count = 0;
				}
			}
		}
		if (repeat_count) {
			if (!--repeat_count) {
				if (repeat_event) {
					put_event(repeat_event);
					repeat_count = 13;
#if 1
					repeat_count = typematic_rate_table[typematic & 31];
#endif
				}
			}
		}

		if (++minuteevent == 60000) {	// 1 minute
			minuteevent = 0;
			if (++kbrefresh >= 9) {		// 9 minute
				kbrefresh = 0;
				kb_enable(false);
				wait_10ms();
				kb_enable(true);
			}
		}
	}

	event = get_event();
	if (event) {
		switch (event >> 8) {
		case EVENT_INIT >> 8:
			break;
		case EVENT_KEYMAKE >> 8:
			c = event & 0xff;
			EncodeKeyCode(c, shiftflags | KEYFLAG_MAKE);
			switch (c) {
			case 44:	shiftflags |= KEYFLAG_SHIFT_L;		break;
			case 57:	shiftflags |= KEYFLAG_SHIFT_R;		break;
			case 58:	shiftflags |= KEYFLAG_CTRL_L;		break;
//			case 64:	shiftflags |= KEYFLAG_CTRL_R;		break;
			case 60:	shiftflags |= KEYFLAG_ALT_L;		break;
//			case 62:	shiftflags |= KEYFLAG_ALT_R;		break;
			case 255:	shiftflags |= KEYFLAG_FN;			break;
			}
			break;
		case EVENT_KEYBREAK >> 8:
			c = event & 0xff;
			EncodeKeyCode(c, shiftflags);
			switch (c) {
			case 44:	shiftflags &= ~KEYFLAG_SHIFT_L;		break;
			case 57:	shiftflags &= ~KEYFLAG_SHIFT_R;		break;
			case 58:	shiftflags &= ~KEYFLAG_CTRL_L;		break;
//			case 64:	shiftflags &= ~KEYFLAG_CTRL_R;		break;
			case 60:	shiftflags &= ~KEYFLAG_ALT_L;		break;
//			case 62:	shiftflags &= ~KEYFLAG_ALT_R;		break;
			case 255:	shiftflags &= ~KEYFLAG_FN;			break;
			}
			break;
		case EVENT_SEND_KB_INDICATOR >> 8:
			break;
		case EVENT_SEND_KB_TYPEMATIC >> 8:
			break;
		}
	}

	c = kb_get();
	if (c >= 0) {
		unsigned short t;
		t = DecodeKeyCode(c);
		if (scan_enable) {
			unsigned short event;
			if (t & 0x8000) {		// break;
				kb_break_delay = 5;
				kb_break_event = (unsigned char)t | EVENT_KEYBREAK;
			} else {						// make
				kb_break_delay = 0;
				event = (unsigned char)t | EVENT_KEYMAKE;
				if (event != repeat_event) {
					put_event(event);
					repeat_event = event;
					repeat_count = 200;
#if 1
					repeat_count = typematic_delay_table[(typematic >> 5) & 3];
#endif
				}
			}
		}
	}

	c = pc_get();
	if (c >= 0) {
		switch (pc_state & 0xff00) {
		case PC_ED:
			indicator = c;
			if (c & 2) {
				shiftflags |= KEYFLAG_NUMLOCK;
			} else {
				shiftflags &= ~KEYFLAG_NUMLOCK;
			}
			LED_SCROLL = (indicator & 1) ? 0 : 1;
			LED_NUM = (indicator & 2) ? 0 : 1;
			LED_CAPS = (indicator & 4) ? 0 : 1;
			pc_put(0xfa);
			pc_state = PC_NORMAL;
			put_event(EVENT_SEND_KB_INDICATOR + 3);		// retry count 3
			break;
		case PC_F0:
			if (c) {
				pc_put(0xfa);
			} else {
				pc_put(2);	// scancode set
			}
			pc_state = PC_NORMAL;
			break;
		case PC_F3:
			typematic = c;
			pc_put(0xfa);
			pc_state = PC_NORMAL;
			put_event(EVENT_SEND_KB_TYPEMATIC + 3);		// retry count 3
			break;
		default:
			switch (c) {
			case 0xed:
				pc_put(0xfa);
				pc_state = PC_ED;
				break;
			case 0xee:
				pc_put(0xee);
				break;
			case 0xf0:
				pc_put(0xfa);
				pc_state = PC_F0;
				break;
			case 0xf1:
				pc_put(0xfa);
				break;
			case 0xf2:
				pc_put(0xfa);
				pc_put(fake_keyboard_id);
				pc_put(fake_keyboard_id >> 8);
				break;
			case 0xf3:
				pc_put(0xfa);
				pc_state = PC_F3;
				break;
			case 0xf4:
			case 0xf5:
			case 0xf6:
			case 0xf7:
			case 0xf8:
			case 0xf9:
			case 0xfa:
			case 0xfb:
			case 0xfc:
			case 0xfd:
				pc_put(0xfa);
				break;
			case 0xfe:
				break;
			case 0xff:
				pc_put(0xfa);
				pc_init_count = 10;		// after 10ms...
				break;
			default:
				pc_put(0xfe);
				break;
			}
		}
	}
}

int main()
{
	PCA0MD = 0x00;

	// initialize system clock
	OSCICN = 0x07;		// oscillator divided by 1
	RSTSRC = 0x04;

	// initialize I/O port
	XBR0 = 0xd7;
	XBR1 = 0x02;
	XBR2 = 0x40;
	P0MDIN = 0xff;
	P0MDOUT = 0x17;

	// initialize timer
	CKCON = 0x00;
	TMOD = 0x20;

	// timer1
	TH1 = -(SYSCLK / 2 / 12 / 9600);		// 9600bps
	TL1 = 0x00;
	TR1 = 1;

	// initialize uart0
	uart0_initialize();

	//

	init_timer2(SYSCLK / 12 / 1000);

	EA = 1;

	//

	qinit(&event_queue);
	qinit(&pc_input_queue);
	qinit(&pc_output_queue);

	pc_set_data_1();
	pc_set_clock_1();

	kb_enable(false);
	{
		LED_NUM = 0; wait_100ms(); LED_NUM = 1;
		LED_CAPS = 0; wait_100ms(); LED_CAPS = 1;
		LED_SCROLL = 0; wait_100ms(); LED_SCROLL = 1;
		LED_CAPS = 0; wait_100ms(); LED_CAPS = 1;
		LED_NUM = 0; wait_100ms(); LED_NUM = 1;
		LED_CAPS = 0; wait_100ms(); LED_CAPS = 1;
		LED_SCROLL = 0; wait_100ms(); LED_SCROLL = 1;
		LED_CAPS = 0; wait_100ms(); LED_CAPS = 1;
		LED_NUM = 0; wait_100ms(); LED_NUM = 1;
	}
	kb_enable(true);

	pc_state = PC_INIT;
//	kb_state = KB_INIT;
	shiftflags = 0;
	pc_init_count = 10;
	indicator = 0;
	typematic = 0x7f;
//	real_keyboard_id = 0;
//	fake_keyboard_id = 0x83ab;
//	break_enable = true;
//	scan_enable = true;
	kb_break_delay = 0;

	while (1) {
		system_handler();
		handler();
	}
	return 0;
}
